#include "ctest1.h"
#include <math.h>

# define timersub(a, b, result)						      \
  do {									      \
    (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;			      \
    (result)->tv_usec = (a)->tv_usec - (b)->tv_usec;			      \
    if ((result)->tv_usec < 0) {					      \
      --(result)->tv_sec;						      \
      (result)->tv_usec += 1000000;					      \
    }									      \
  } while (0)

char configFileName[MAX_LEN_CONFIG_FILE];
char fullConfigFileName[MAX_LEN_FULL_CONFIG_FILE];
struct TParamCTest1 params;
struct TChannel* channels;
struct TDevice* devices;

int main(int argc, char* argv[])
{
    initLogCTest1();
    LOGGER_PRINT("%s %s\n","ctest1 - version", VERSION_APP);
    int ret = GOOD_CODE;
    if (!(argc == 2 || argc == 3)) { // проверка кол-ва аргументов
        usage:
#ifdef HAS_EXTA429_LIBRARY
        LOGGER_PRINT("%s\n","Usage: ctest1 <configfilename> <possible external device IP address:port>\n" \
                            "Example: ctest1 ctest1.txt 192.168.12.34:19191\n");
#endif // HAS_EXTA429_LIBRARY

        LOGGER_PRINT("%s\n","Usage: ctest1 <configfilename>\n" \
                            "Example: ctest1 ctest1.txt\n");

        ret = BAD_CODE;
    } else { // кол-во аргументов - ок
        if (sscanf(argv[1], "%s", configFileName) != 1) { // чтение имени конфигурационного файла
          LOGGER_PRINT("%s\n", "invalid configfilename\n");
          goto usage;
        } else { //имя конфигурационного файла - ок
          if(argc == 3)
          {
#ifdef HAS_EXTA429_LIBRARY
              if(!setNetworkDevice(argv[2])) // при наличии, парсим ip адрес
#endif
                    goto usage;
          }

          constructFullConfigFileName(fullConfigFileName, MAX_LEN_FULL_CONFIG_FILE, configFileName);
          initDefaultParamsCTest1(&params);
          ret = readConfigFile(fullConfigFileName, &params); // чтение параметров из конфигурационного файла
          if (ret == BAD_CODE) { //параметры из конфигурационного файла - error
              LOGGER_PRINT("Can not read test params from '%s' \n", fullConfigFileName);
          } else { //параметры из конфигурационного файла - ок
              procCTest1(params);
          }
        }
    }
    LOGGER_FLUSH();
    LOGGER_CLOSE();

    close_all_abonents();
    return ret;
}

void constructFullConfigFileName(char* fullname, int size, char* name) {
    snprintf(fullname, size, "%s%s", STR_CONFIG, name);
}

int readConfigFile(char* fullname, struct TParamCTest1* params) {
    FILE* file = fopen(fullname, "r");
    char* str;
    if (file == NULL) {
        LOGGER_PRINT("Can not open config file '%s'\n", fullname);
        return BAD_CODE;
    }
    LOGGER_PRINT("parsing '%s':\n", fullname);
    while (!feof(file)) {
        readline(file, &str);
        LOGGER_PRINT("%s\n", str); // for debug
        char typeArgStr[1];
        int valArg = BAD_CODE;
        sscanf(str, "%s %d", typeArgStr, &valArg);
        if (valArg == BAD_CODE)
            continue;
        switch (typeArgStr[0]) {
            case T_CTEST1_SPEED:    params->mSpeed = valArg;        break;
            case T_CTEST1_PARITY:   params->mParityBit = valArg;    break;
            case T_CTEST1_REVERCE:  params->mReverce = valArg;      break;
            case T_CTEST1_MODE:     params->mMode = valArg;         break;
            case T_CTEST1_TIMEOUT:  params->mPrintTimeout = valArg; break;
            case T_CTEST1_CNT_MSG:  params->mCountMessages = valArg; break;
        }
    }
    fclose(file);
    LOGGER_PRINT("PARAMS: speed= %d, parity= %d, reverce= %d, mode= %d, timeout= %d, count messages= %d\n\n", \
           params->mSpeed, params->mParityBit, params->mReverce, params->mMode, params->mPrintTimeout, params->mCountMessages);
    return GOOD_CODE;
}

void procCTest1(struct TParamCTest1 params) {
    struct timeval tvStart, tvCur, tvDiff;
    int countDevices = 0;
    gettimeofday(&tvStart, NULL);
    getArrayDevices(&devices, &countDevices);
    LOGGER_PRINT("Has founded %d devices \n\n", countDevices);
    int countChannels = 0;
    getArrayChannels(&channels, &countChannels);
    LOGGER_PRINT("Has founded %d channels \n\n", countChannels);
    int counters[countChannels];    // счётчики данных блоков ДМА
    // инициализация девайсов
    for (int index = 0; index < countDevices; index++) {
        Abonent abonent = open_abonent_from_string(devices[index].mNameChannel);
        IS_EXT_WRAPPER(abonent, a429_clearDma);
        IS_EXT_WRAPPER(abonent, a429_manageDma, A429_ON);
        IS_EXT_WRAPPER(abonent, a429_manageInterruptMask,
                                 1<<8 | 1<<9 | 1<<10 | 1<<11 | 1<<12 | 1<<13 | 1<<14 | 1<<15,
                                 1<<8 | 1<<9 | 1<<10 | 1<<11 | 1<<12 | 1<<13 | 1<<14 | 1<<15);
        IS_EXT_WRAPPER(abonent, a429_clearScIntMask);
        IS_EXT_WRAPPER(abonent, a429_manageScIntMask, A429_IN_RK_1, A429_SC_RISE, A429_ON);
        IS_EXT_WRAPPER(abonent, a429_manageScIntMask, A429_IN_RK_2, A429_SC_RISE, A429_ON);
        IS_EXT_WRAPPER(abonent, a429_manageScIntMask, A429_IN_RK_3, A429_SC_RISE, A429_ON);
        IS_EXT_WRAPPER(abonent, a429_manageScIntMask, A429_IN_RK_4, A429_SC_RISE, A429_ON);
        IS_EXT_WRAPPER(abonent, a429_manageScIntMask, A429_IN_RK_5, A429_SC_RISE, A429_ON);
        IS_EXT_WRAPPER(abonent, a429_manageScIntMask, A429_IN_RK_6, A429_SC_RISE, A429_ON);
        IS_EXT_WRAPPER(abonent, a429_manageScIntMask, A429_IN_RK_7, A429_SC_RISE, A429_ON);
        IS_EXT_WRAPPER(abonent, a429_manageScIntMask, A429_IN_RK_8, A429_SC_RISE, A429_ON);
        T_TRACKING_INT trackInt;
        // включаем слежение за РК ВХ
        trackInt.manage = T_TRACKING_INT_MAN_ON;
        IS_EXT_WRAPPER(abonent, a429_manageTrackingInterrupt, &trackInt);
    }
    for (int index = 0; index < countChannels; index++) {
        Abonent abonent = open_abonent_from_string(channels[index].mNameChannel);
        IS_EXT_WRAPPER(abonent, a429_deinit);
    }
    // инициализация каналов
    for (int index = 0; index < countChannels; index++) {
        counters[index] = 0;
        Abonent abonent = open_abonent_from_string(channels[index].mNameChannel);
        IS_EXT_WRAPPER(abonent, a429_manageEnBit, A429_OFF);
        IS_EXT_WRAPPER(abonent, a429_manageChannelDma, A429_ON);
        IS_EXT_WRAPPER(abonent, a429_setSpeed, params.mSpeed);
        IS_EXT_WRAPPER(abonent, a429_parityManage, A429_ON, params.mParityBit);
        IS_EXT_WRAPPER(abonent, a429_manageReverce, params.mReverce);
        if (channels[index].mInfo.typeChannel == T_INFO_CHANNEL_TYPE_TRANSMITTER) {
            IS_EXT_WRAPPER(abonent, a429_txSetMode, params.mMode);
            IS_EXT_WRAPPER(abonent, a429_txGapBits, 4);
        }        
        IS_EXT_WRAPPER(abonent, a429_manageEnBit, A429_ON);
        if (channels[index].mInfo.typeChannel == T_INFO_CHANNEL_TYPE_TRANSMITTER) {  
            IS_EXT_WRAPPER(abonent, a429_txStartStop, A429_START);
        } else {
            IS_EXT_WRAPPER(abonent, a429_rxSetReceiveReadyByRkIn, RKIN_NUM_1);
        }
    }
    LOGGER_PRINT("%s\n", "test is processing...");
    // выполнение алгоритма тестирования
    for(;;) {
        for (int index = 0; index < countDevices; index++) {
            Abonent abonent = open_abonent_from_string(devices[index].mNameChannel);
            IS_EXT_WRAPPER(abonent, a429_manageScOut, A429_SC_OUT_1, A429_SC_OUT_1, A429_SC_OUT_1, A429_SC_OUT_1, A429_SC_OUT_1, A429_SC_OUT_1, A429_SC_OUT_1, A429_SC_OUT_1);
        }
        for (int index = 0; index < countChannels; index++) {
            T_CONTAINER_DMA dma;
            dma.count = 0;
            // чтение данных
            Abonent abonent = open_abonent_from_string(channels[index].mNameChannel);
            IS_EXT_WRAPPER(abonent, a429_readDma, &dma);

            counters[index] += dma.count;
            printErrA429Rcv(&channels[index], &dma);
        }
        gettimeofday(&tvCur, NULL);
        timersub(&tvCur, &tvStart, &tvDiff);    // вычисляем время прошедшее с предыдущего вывода результатов
        if (tvDiff.tv_sec * 1000000 + tvDiff.tv_usec >= (params.mPrintTimeout /*seconds*/ * 1000000)) { //проверка таймаута
            for (int index = 0; index < countDevices; index++) {
                Abonent abonent = open_abonent_from_string(devices[index].mNameChannel);
                IS_EXT_WRAPPER(abonent, a429_manageScOut, A429_SC_OUT_0, A429_SC_OUT_0, A429_SC_OUT_0, A429_SC_OUT_0, A429_SC_OUT_0, A429_SC_OUT_0, A429_SC_OUT_0, A429_SC_OUT_0);
            }
            // выдача информации на экран и в лог и сброс накопленных счётчиков принятых/переданных данных
            for (int index = 0; index < countChannels; index++) {
                LOGGER_PRINT("Channel '%s' = %d blocks \n", channels[index].mNameChannel, counters[index]);
                counters[index] = 0;
            }
            for (int index = 0; index < countChannels; index++) {
                if (channels[index].mInfo.typeChannel == T_INFO_CHANNEL_TYPE_TRANSMITTER) {
                    Abonent abonent = open_abonent_from_string(channels[index].mNameChannel);
                    // отправка данных
                    for (int i=0; i<params.mCountMessages; i++) {
                        IS_EXT_WRAPPER(abonent, a429_writeTxFifo, 0xAAAAAAAA);
                    }
                }
            }
            for (int index = 0; index < countDevices; index++) {
                Abonent abonent = open_abonent_from_string(devices[index].mNameChannel);
                T_INFO_DEVICE info;
                IS_EXT_WRAPPER(abonent, a429_getDevInfoChannels, &info);
                int cntRk = max(2, min(info.countRecieverChannels, info.countTransmitterChannels) / 2);

                T_CHECK_TRACKING_INT trackCheckInt;
                // проверяем результат ВХ РК
                IS_EXT_WRAPPER(abonent, a429_checkTrackingInterrupt, &trackCheckInt);
                LOGGER_PRINT("Device '%s' devId= %d - checking", devices[index].mNameChannel, devices[index].mDevId);

                for (int i=0; i < cntRk; i++) {
                    LOGGER_PRINT(" IN_RK %d= %d;", i+1, trackCheckInt.rkInAppearCount[i]);
                }

                LOGGER_PRINT("%s", "\n");
            }
            for (int index = 0; index < countDevices; index++) {
                Abonent abonent = open_abonent_from_string(devices[index].mNameChannel);
                T_TRACKING_INT trackInt;
                trackInt.manage = T_TRACKING_INT_MAN_OFF;
                IS_EXT_WRAPPER(abonent, a429_manageTrackingInterrupt, &trackInt);
                trackInt.manage = T_TRACKING_INT_MAN_ON;
                IS_EXT_WRAPPER(abonent, a429_manageTrackingInterrupt, &trackInt);
            }
            tvStart = tvCur;
            LOGGER_PRINT("-has passed %d seconds-\n", params.mPrintTimeout);
        }
        LOGGER_FLUSH();
    }
}

void initLogCTest1(void) {
    char filename[LOGGER_FILENAME_LENGTH];
    constructLogFilename(filename, LOGGER_FILENAME_LENGTH, "ctest1");
    LOGGER_OPEN(filename);
}

void initDefaultParamsCTest1(struct TParamCTest1* params) {
    params->mSpeed = 2;
    params->mParityBit = 1;
    params->mReverce = 0;
    params->mMode = 0;
    params->mPrintTimeout = 3;
    params->mCountMessages = 50;;
}

