#include "ctest2.h"

# define timersub(a, b, result)						      \
  do {									      \
    (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;			      \
    (result)->tv_usec = (a)->tv_usec - (b)->tv_usec;			      \
    if ((result)->tv_usec < 0) {					      \
      --(result)->tv_sec;						      \
      (result)->tv_usec += 1000000;					      \
    }									      \
  } while (0)

char configFileName[MAX_LEN_CONFIG_FILE];
char fullConfigFileName[MAX_LEN_FULL_CONFIG_FILE];
struct TParamCTest2 params;
struct TChannel* channels;
struct TDevice* devices;

int main(int argc, char* argv[])
{
    initLogCTest2();
    LOGGER_PRINT("%s %s\n","ctest2 - version", VERSION_APP);
    int ret = GOOD_CODE;
    if (!(argc == 2 || argc == 3)) { // проверка кол-ва аргументов
        usage:

#ifdef HAS_EXTA429_LIBRARY
        LOGGER_PRINT("%s\n","Usage: ctest2 <configfilename> <possible external device IP address:port>\n" \
                            "Example: ctest2 ctest2.txt 192.168.12.34:19191\n");
#endif // HAS_EXTA429_LIBRARY

        LOGGER_PRINT("%s\n","Usage: ctest2 <configfilename>\n" \
                            "Example: ctest2 ctest2.txt\n");

        ret = BAD_CODE;
    } else { // кол-во аргументов - ок
        if (sscanf(argv[1], "%s", configFileName) != 1) { // чтение имени конфигурационного файла
          LOGGER_PRINT("%s\n", "invalid configfilename\n");
          goto usage;
        } else { //имя конфигурационного файла - ок
            if(argc == 3)
            {
  #ifdef HAS_EXTA429_LIBRARY
                if(!setNetworkDevice(argv[2])) // при наличии, парсим ip адрес
  #endif
                      goto usage;
            }

          constructFullConfigFileName(fullConfigFileName, MAX_LEN_FULL_CONFIG_FILE, configFileName);
          ret = readConfigFile(fullConfigFileName, &params); // чтение параметров из конфигурационного файла
          if (ret == BAD_CODE) { //параметры из конфигурационного файла - error
              LOGGER_PRINT("Can not read test params from '%s' \n", fullConfigFileName);
          } else { //параметры из конфигурационного файла - ок
              procCTest2(params);
          }
        }
    }
    LOGGER_FLUSH();
    LOGGER_CLOSE();

    close_all_abonents();
    return ret;
}

void constructFullConfigFileName(char* fullname, int size, char* name) {
    snprintf(fullname, size, "%s%s", STR_CONFIG, name);
}

int readConfigFile(char* fullname, struct TParamCTest2* params) {
    FILE* file = fopen(fullname, "r");
    char* str;
    if (file == NULL) {
        LOGGER_PRINT("Can not open config file '%s'\n", fullname);
        return BAD_CODE;
    }
    LOGGER_PRINT("parsing '%s':\n", fullname);
    while (!feof(file)) {
        readline(file, &str);
        LOGGER_PRINT("%s\n", str); // for debug
        char typeArgStr[1];
        int valArg = BAD_CODE;
        sscanf(str, "%s %d", typeArgStr, &valArg);
        if (valArg == BAD_CODE)
            continue;
        switch (typeArgStr[0]) {
            case T_CTEST2_SPEED:    params->mSpeed = valArg;        break;
            case T_CTEST2_PARITY:   params->mParityBit = valArg;    break;
            case T_CTEST2_REVERCE:  params->mReverce = valArg;      break;
            case T_CTEST2_TIMEOUT:  params->mPrintTimeout = valArg; break;
        }
    }
    fclose(file);
    LOGGER_PRINT("PARAMS: speed= %d, parity= %d, reverce= %d, timeout= %d\n\n", \
           params->mSpeed, params->mParityBit, params->mReverce, params->mPrintTimeout);
    return GOOD_CODE;
}

#define WAIT_SEC(sec) \
    for(;;){ \
        gettimeofday(&tvCur, NULL); \
        timersub(&tvCur, &tvStart, &tvDiff); \
        if (tvDiff.tv_sec * 1000000 + tvDiff.tv_usec >= (sec /*seconds*/ * 1000000)) { \
            tvStart = tvCur; \
            break; \
        } \
    }

#define OPEN_CH(index) \
    Abonent abonent = open_abonent_from_string(channels[index].mNameChannel);

#define CLOSE_CH(index)

#define OPEN_DEV(index) \
    Abonent abonent = open_abonent_from_string(devices[index].mNameChannel);

#define CLOSE_DEV(index)

void procCTest2(struct TParamCTest2 params) {
    RAM_BLOCK desc, data;
    loadMemCTest2(&desc);
    constructMemDataCTest2(&data);
    struct timeval tvStart, tvCur, tvDiff;
    int countDevices = 0;
    gettimeofday(&tvStart, NULL);
    getArrayDevices(&devices, &countDevices);
    LOGGER_PRINT("Has founded %d devices \n\n", countDevices);
    int countChannels = 0;
    getArrayChannels(&channels, &countChannels);
    LOGGER_PRINT("Has founded %d channels \n\n", countChannels);
    int counters[countChannels];    // счётчики данных блоков ДМА
    int errors[countChannels];
    // выкл enbit
    for (int index = 0; index < countChannels; index++) {
        OPEN_CH(index);
        IS_EXT_WRAPPER(abonent, a429_txStartStop, A429_STOP);
        IS_EXT_WRAPPER(abonent, a429_manageEnBit, A429_OFF);
        CLOSE_CH(index);
    }
    // сброс каналов ram cls
    for (int index = 0; index < countChannels; index++) {
        OPEN_CH(index);
        IS_EXT_WRAPPER(abonent, a429_deinit);
        IS_EXT_WRAPPER(abonent, a429_ramCls);
        CLOSE_CH(index);
    }
    // сбрасываем дма
    for (int index = 0; index < countDevices; index++) {
        OPEN_DEV(index);
        IS_EXT_WRAPPER(abonent, a429_manageDma, A429_OFF);
        IS_EXT_WRAPPER(abonent, a429_clearDma);
        CLOSE_DEV(index);
    }
    // инициализация девайсов - вкл дма
    for (int index = 0; index < countDevices; index++) {
        OPEN_DEV(index);
        IS_EXT_WRAPPER(abonent, a429_manageDma, A429_ON);
        CLOSE_DEV(index);
    }
    for (int index = 0; index < countChannels; index++) {
        OPEN_CH(index);
        IS_EXT_WRAPPER(abonent, a429_deinit);
        CLOSE_CH(index);
    }
    // инициализация каналов
    for (int index = 0; index < countChannels; index++) {
        counters[index] = 0;
        errors[index] = 0;
        OPEN_CH(index);
        IS_EXT_WRAPPER(abonent, a429_manageEnBit, A429_OFF);
        IS_EXT_WRAPPER(abonent, a429_manageChannelDma, A429_ON);
        IS_EXT_WRAPPER(abonent, a429_setSpeed, params.mSpeed);
        IS_EXT_WRAPPER(abonent, a429_parityManage, A429_ON, params.mParityBit);
        IS_EXT_WRAPPER(abonent, a429_manageReverce, params.mReverce);
        if (channels[index].mInfo.typeChannel == T_INFO_CHANNEL_TYPE_TRANSMITTER) {
            IS_EXT_WRAPPER(abonent, a429_txSkipRrt, A429_ON);
            IS_EXT_WRAPPER(abonent, a429_txSetMode, A429_MODE_1);
            IS_EXT_WRAPPER(abonent, a429_txGapBits, 4);
            // прошивка в передатчик mem файлов
            IS_EXT_WRAPPER(abonent, a429_writeRam, &desc);
            IS_EXT_WRAPPER(abonent, a429_writeRam, &data);
        }        
        CLOSE_CH(index);
    }
    for (int index = 0; index < countChannels; index++) {
        OPEN_CH(index);
        IS_EXT_WRAPPER(abonent, a429_manageEnBit, A429_ON);
        CLOSE_CH(index);
    }
    LOGGER_PRINT("%s\n", "test is processing...");
    for (int index = 0; index < countChannels; index++) {
        OPEN_CH(index);
        if (channels[index].mInfo.typeChannel == T_INFO_CHANNEL_TYPE_TRANSMITTER) {
            IS_EXT_WRAPPER(abonent, a429_txStartStop, A429_START);
        }
        CLOSE_CH(index);
    }
    // выполнение алгоритма тестирования
    for(;;) {
        for (int index = 0; index < countChannels; index++) {
            T_CONTAINER_DMA dma;
            dma.count = 0;
            OPEN_CH(index);
            // чтение данных
            IS_EXT_WRAPPER(abonent, a429_readDma, &dma);
            T_INFO_CHANNEL info;
            IS_EXT_WRAPPER(abonent, a429_getChannelInfo, &info);
            CLOSE_CH(index);
            counters[index] += dma.count;
            if (T_INFO_CHANNEL_TYPE_RECIEVER == info.typeChannel) {
                for (int indexBlk=0; indexBlk<dma.count; indexBlk++) {
                    if (dma.blocks[indexBlk].word2 & ( 1<<21 | 1<<22 | 1<<23)) {
                       errors[index] += 1;
                    }
                }
            }
            printErrA429Rcv(&channels[index], &dma);
        }
        gettimeofday(&tvCur, NULL);
        timersub(&tvCur, &tvStart, &tvDiff);    // вычисляем время прошедшее с предидущего вывода результатов
        if (tvDiff.tv_sec * 1000000 + tvDiff.tv_usec >= (params.mPrintTimeout /*seconds*/ * 1000000)) { //проверка таймаута
            // выдача информации на экран и в лог и сброс накопленных счётчиков принятых/переданных данных
            for (int index = 0; index < countChannels; index++) {
                LOGGER_PRINT("Channel '%s' = %d blocks; %d errors \n", channels[index].mNameChannel, counters[index], errors[index]);
                counters[index] = 0;
                errors[index] = 0;
            }
            tvStart = tvCur;
            LOGGER_PRINT("-has passed %d seconds-\n", params.mPrintTimeout);
            for (int index = 0; index < countChannels; index++) {
                OPEN_CH(index);
                if (channels[index].mInfo.typeChannel == T_INFO_CHANNEL_TYPE_TRANSMITTER) {
                    IS_EXT_WRAPPER(abonent, a429_writeRam, &data);
                }
                CLOSE_CH(index);
            }
        }
        LOGGER_FLUSH();
    }
}

void initLogCTest2(void) {
    char filename[LOGGER_FILENAME_LENGTH];
    constructLogFilename(filename, LOGGER_FILENAME_LENGTH, "ctest2");
    LOGGER_OPEN(filename);
}

void loadMemCTest2(RAM_BLOCK* ram) {
    ram->length = 256;
    ram->shift = 0;
    ram->type = RAM_BLOCK_TYPE_DESCRIPTOR;
    ram->dwords = malloc(sizeof(unsigned int) * ram->length);
    char filename[MEM_FILENAME_LENGTH];
    snprintf(filename, MEM_FILENAME_LENGTH, "%s%s", STR_MEM, "desk_test_consol.mem");
    FILE* file = fopen(filename, "r");
    char* str;
    if (file == NULL) {
        LOGGER_PRINT("Can not open config file '%s'\n", filename);
        return;
    }
    int index = 0;
    while (!feof(file)) {
        if (index > 255)
            break;
        readline(file, &str);
        char num[10];
        unsigned int word = 0;
        sscanf(str, "%s %x", num, &word);
        LOGGER_PRINT("%s - readed 0x%x\n", str, word); // for debug
        ram->dwords[index] = word;
        index++;
    }
    fclose(file);
}

void constructMemDataCTest2(RAM_BLOCK* ram) {
    ram->length = 256;
    ram->shift = 0;
    ram->type = RAM_BLOCK_TYPE_DATA;
    ram->dwords = malloc(sizeof(unsigned int) * ram->length);
    for (int i=0; i<ram->length; i++) {
        ram->dwords[i] = 0xAAAAAAAA;
    }
}

