#include "global.h"

FILE* LOGGER_FILE;

#ifdef HAS_EXTA429_LIBRARY
const char *g_networkDevice = NULL;
char g_networkIP[30];
int g_networkPort = 0;

bool setNetworkDevice(const char *device)
{
    int ip[4];

    if(sscanf(device, "%d.%d.%d.%d:%d", &ip[0], &ip[1], &ip[2], &ip[3], &g_networkPort) != 5)
        return false;

    for(int i = 0; i < 4; ++i)
        if(ip[i] < 0 || ip[i] > 255)
            return false;

    if(g_networkPort < 0 || g_networkPort > 0xFFFF)
        return false;

    sprintf(g_networkIP, "%d.%d.%d.%d", ip[0], ip[1], ip[2], ip[3]);
    g_networkDevice = device;
    return true;
}
#endif

int getCountDevices(void)
{
#ifdef HAS_EXTA429_LIBRARY
    if(g_networkDevice != NULL)
        return 1;
#endif

    int count = 0;
    DIR* dir = opendir(STR_DEV);
    if (dir) {
        struct dirent *direntries;
        while (((direntries = readdir(dir)) != NULL)) {
            char* isFindedChannelPCI = strstr(direntries->d_name, STR_DEV_PCI_PATTERN);
            char* isFindedChannelUSB = strstr(direntries->d_name, STR_DEV_USB_PATTERN);
            char* isFindedChannel0 = strstr(direntries->d_name, STR_CH0_PATTERN);
            if ((isFindedChannelPCI != NULL || isFindedChannelUSB != NULL) && (isFindedChannel0 != NULL)) {
//                printf("device: %s\n", direntries->d_name);
                count++;
            }
        }
        closedir(dir);
    }
    return count;
}

int getCountChannels(void)
{
#ifdef HAS_EXTA429_LIBRARY
    if(g_networkDevice != NULL)
    {
        char strFirstChannel[50];
        sprintf(strFirstChannel, "%s:%d:0", g_networkIP, g_networkPort);
        Abonent abonent = open_abonent_from_string(strFirstChannel);
        VERSION ver;
        T_INFO_DEVICE devInfo;
        IS_EXT_WRAPPER(abonent, a429_getDeviceInfo, &ver);
        if(GOOD == IS_EXT_WRAPPER(abonent, a429_getDevInfoChannels, &devInfo))
        {
            // debug
            LOGGER_PRINT("%s - devId %d - revision %d - приёмники %d (шт), передатчики %d (шт)\n", strFirstChannel, ver.device_id, ver.revision, devInfo.countRecieverChannels, devInfo.countTransmitterChannels);
            return devInfo.countRecieverChannels + devInfo.countTransmitterChannels;
        }

        return 0;
    }
#endif

    int count = 0;
    DIR* dir = opendir(STR_DEV);
    if (dir) {
        struct dirent *direntries;
        while (((direntries = readdir(dir)) != NULL)) {
            char* isFindedChannelPCI = strstr(direntries->d_name, STR_DEV_PCI_PATTERN);
            char* isFindedChannelUSB = strstr(direntries->d_name, STR_DEV_USB_PATTERN);

            if (isFindedChannelPCI != NULL || isFindedChannelUSB != NULL) {
//                printf("%s\n", direntries->d_name);
                count++;
            }
        }
        closedir(dir);
    }
    return count;
}

void requestDeviceInfo(struct TDevice** devices, int index)
{
    Abonent abonent = open_abonent_from_string((*devices)[index].mNameChannel);
    VERSION ver;
    T_INFO_DEVICE devInfo;
    IS_EXT_WRAPPER(abonent, a429_getDeviceInfo, &ver);
    (*devices)[index].mDevId = ver.device_id;
    IS_EXT_WRAPPER(abonent, a429_getDevInfoChannels, &devInfo);
    // debug
    LOGGER_PRINT("%s - devId %d - приёмники %d (шт), передатчики %d (шт)\n", (*devices)[index].mNameChannel, (*devices)[index].mDevId, devInfo.countRecieverChannels, devInfo.countTransmitterChannels);
}

void getArrayDevices(struct TDevice** devices, int* length)
{
    *length = getCountDevices();
    *devices = malloc(sizeof(struct TDevice) * *length);
    int index = 0;

#ifdef HAS_EXTA429_LIBRARY
    if(g_networkDevice != NULL)
    {
        snprintf((*devices)[index].mNameChannel, CHANNEL_NAME_LENGTH, "%s:%d:0", g_networkIP, g_networkPort);
        requestDeviceInfo(devices, index);
        return;
    }
#endif

    DIR* dir = opendir(STR_DEV);
    if (dir) {
        struct dirent *direntries;
        while (((direntries = readdir(dir)) != NULL)) {
            char* isFindedChannelPCI = strstr(direntries->d_name, STR_DEV_PCI_PATTERN);
            char* isFindedChannelUSB = strstr(direntries->d_name, STR_DEV_USB_PATTERN);

            char* isFindedChannel0 = strstr(direntries->d_name, STR_CH0_PATTERN);
            if ((isFindedChannelPCI != NULL || isFindedChannelUSB != NULL) && (isFindedChannel0 != NULL)) {
                snprintf((*devices)[index].mNameChannel, CHANNEL_NAME_LENGTH, "%s/%s", STR_DEV, direntries->d_name);
                requestDeviceInfo(devices, index);
                index++;
            }
        }
        closedir(dir);
    }
}

void requestChannelInfo(struct TChannel **channels, int index)
{
    Abonent abonent = open_abonent_from_string((*channels)[index].mNameChannel);
    IS_EXT_WRAPPER(abonent, a429_getChannelInfo, &((*channels)[index].mInfo));
    // debug
    if ((*channels)[index].mInfo.typeChannel == T_INFO_CHANNEL_TYPE_RECIEVER) {
        LOGGER_PRINT("%s - num %d - T_INFO_CHANNEL_TYPE_RECIEVER\n", (*channels)[index].mNameChannel, (*channels)[index].mInfo.numberChannel);
    } else {
        LOGGER_PRINT("%s - num %d - T_INFO_CHANNEL_TYPE_TRANSMITTER\n", (*channels)[index].mNameChannel, (*channels)[index].mInfo.numberChannel);
    }
}

void getArrayChannels(struct TChannel **channels, int *length)
{
    *length = getCountChannels();
    *channels = malloc(sizeof(struct TChannel) * *length);
    int index = 0;

#ifdef HAS_EXTA429_LIBRARY
    if(g_networkDevice != NULL)
    {
        while(index < *length)
        {
            snprintf((*channels)[index].mNameChannel, CHANNEL_NAME_LENGTH, "%s:%d:%d", g_networkIP, g_networkPort, index);
            requestChannelInfo(channels, index);
            index++;
        }
        return;
    }
#endif

    DIR* dir = opendir(STR_DEV);
    if (dir) {
        struct dirent *direntries;
        while (((direntries = readdir(dir)) != NULL)) {
            char* isFindedChannelPCI = strstr(direntries->d_name, STR_DEV_PCI_PATTERN);
            char* isFindedChannelUSB = strstr(direntries->d_name, STR_DEV_USB_PATTERN);

            if (isFindedChannelPCI != NULL || isFindedChannelUSB != NULL) {
                snprintf((*channels)[index].mNameChannel, CHANNEL_NAME_LENGTH, "%s/%s", STR_DEV, direntries->d_name);
                requestChannelInfo(channels, index);
                index++;
            }
        }
        closedir(dir);
    }
}

void _readline(FILE* stream, char** line, int* count) {
    int ch = fgetc(stream);
    if ((ch == '\n') || (ch == EOF)) {
//        printf("end line - %d\n", *count);
        *line = malloc(*count + 1);
        (*line)[*count] = 0;
//        printf("down to up\n");
        return;
    } else {
        (*count)++;
//        printf("down - %d\n", *count);
        _readline(stream, line, count);
//        printf("up - %d\n", *count);
        (*line)[--(*count)] = ch;
    }
}


void readline(FILE* stream, char** line) {
    int count = 0;
    _readline(stream, line, &count);
}

void constructLogFilename(char* filename, int length, char* testname) {
    time_t rawtime;
    struct tm * timeinfo;
    time(&rawtime);
    timeinfo = localtime(&rawtime);
    snprintf(filename, length, "%s%s-%dy-%dm-%dd-%dh-%dm-%ds.txt", STR_LOGS, testname, \
             timeinfo->tm_year + 1900, timeinfo->tm_mon + 1, timeinfo->tm_mday, timeinfo->tm_hour, timeinfo->tm_min, timeinfo->tm_sec);
}

int hasDmaBlockA429RcvErr(T_BLOCK_DMA* block) {
    if (block->word2 & (1<<21)) {
        return 1;
    }
    if (block->word2 & (1<<22)) {
        return 1;
    }
    if (block->word2 & (1<<23)) {
        return 1;
    }
    return 0;
}

void printErrA429Rcv(struct TChannel* ch, T_CONTAINER_DMA* dma) {
    if (ch->mInfo.typeChannel == T_INFO_CHANNEL_TYPE_RECIEVER) {
        for (int i=0; i < dma->count; i++) {
            if (hasDmaBlockA429RcvErr( &dma->blocks[i] )) {
                LOGGER_PRINT("Error in A429 %s dma word2= 0x%x:\n", ch->mNameChannel, dma->blocks[i].word2);
                if (dma->blocks[i].word2 & (1<<21)) {
                    LOGGER_PRINT("%s;\n", "Ошибка кодировки бита");
                }
                if (dma->blocks[i].word2 & (1<<22)) {
                    LOGGER_PRINT("%s;\n", "Ошибка паузы");
                }
                if (dma->blocks[i].word2 & (1<<23)) {
                    LOGGER_PRINT("%s;\n", "Ошибка чётности");
                }
            }
        }
    }
}


struct AbonentItem *g_abonentCache = NULL;

void close_all_abonents()
{
    struct AbonentItem *tmp;

    while(g_abonentCache != NULL)
    {
        if(g_abonentCache->abonent.isExt)
        {
#ifdef HAS_EXTA429_LIBRARY
            exta429_close(g_abonentCache->abonent.extFd);
#endif
        }
        else
        {
#ifdef HAS_A429_LIBRARY
            a429_close(g_abonentCache->abonent.fd);
#endif
        }

        tmp = g_abonentCache->next;
        free(g_abonentCache);
        g_abonentCache = tmp;
    }
}

Abonent open_abonent_from_string(char *path)
{
    int ipAddr[4], port, channel;
    Abonent result;

    for(struct AbonentItem *abonentCacheTmp = g_abonentCache; abonentCacheTmp != NULL; abonentCacheTmp = abonentCacheTmp->next)
        if(0 == strncmp(path, abonentCacheTmp->abonent.name, CHANNEL_NAME_LENGTH))
            return abonentCacheTmp->abonent;

    memset(&result, 0, sizeof(Abonent));
    result.isExt = false;

#ifdef HAS_A429_LIBRARY
    result.fd = -1;
#endif

#ifdef HAS_EXTA429_LIBRARY
    result.extFd.socketFD = -1;
#endif

    strncpy(result.name, path, CHANNEL_NAME_LENGTH);

    if(6 == sscanf(path, "%d.%d.%d.%d:%d:%d", &ipAddr[0], &ipAddr[1], &ipAddr[2], &ipAddr[3], &port, &channel))
    {
#ifdef HAS_EXTA429_LIBRARY
        char str_ipAddr[100];
        sprintf(str_ipAddr, "%d.%d.%d.%d", ipAddr[0], ipAddr[1], ipAddr[2], ipAddr[3]);

        result.isExt = true;
        result.extFd = exta429_open(str_ipAddr, port, channel);
#endif
    }
    else
    {
#ifdef HAS_A429_LIBRARY
        // локальное устройство
        result.fd = a429_open(path);
#endif
    }

    {
        struct AbonentItem *nextAbonentItem;

        if(g_abonentCache == NULL)
        {
            g_abonentCache = (struct AbonentItem *) malloc(sizeof(struct AbonentItem));
            nextAbonentItem = g_abonentCache;
        }
        else
        {
            nextAbonentItem = g_abonentCache;

            while(nextAbonentItem->next != NULL)
                nextAbonentItem = nextAbonentItem->next;

            nextAbonentItem->next = (struct AbonentItem *) malloc(sizeof(struct AbonentItem));
            nextAbonentItem = nextAbonentItem->next;
        }

        memset(nextAbonentItem, 0, sizeof(struct AbonentItem));
        nextAbonentItem->abonent = result;
    }

    return result;
}
